def partition(func, values):
    true = []
    false = []

    for value in values:
        if func(value):
            true.append(value)
        else:
            false.append(value)

    return true, false


def single(values):
    if len(values) == 1:
        return values[0]
    else:
        raise Exception("Expected 1 but got {}".format(len(values)))


def find(predicate, values):
    for value in values:
        if predicate(value):
            return value


def unique(values, key):
    result = []
    seen = set()

    for value in values:
        key_value = key(value)
        if key_value not in seen:
            seen.add(key_value)
            result.append(value)

    return result



import abc
from functools import partial
from itertools import groupby

from attr import assoc
from graphql import GraphQLError, GraphQLField, GraphQLObjectType, GraphQLList, GraphQLSchema
from graphql.execution import execute as graphql_execute, ExecutionResult
from graphql.language.parser import parse
from graphql.validation import validate
import six

from .requests import request_from_graphql_ast, request_from_graphql_document, Request, field_key
from .util import partition, unique


def executor(root):
    schema = GraphQLSchema(query=root.to_graphql_type())
    return partial(_execute, schema, root)


def execute(root, *args, **kwargs):
    return executor(root)(*args, **kwargs)


def _execute(schema, root, query, context=None, variables=None):
    try:
        schema = GraphQLSchema(query=root.to_graphql_type())
        ast = parse(query)
        validation_errors = validate(schema, ast)
        if validation_errors:
            return ExecutionResult(
                errors=validation_errors,
                invalid=True,
            )

        request = request_from_graphql_document(ast, root, context=context, variables=variables)
        data = root.fetch(request.query, None)[0].value

        if request.schema_query is not None:
            schema_result = graphql_execute(
                schema,
                request.schema_query,
                variable_values=variables,
            )
            if schema_result.invalid:
                return schema_result

            data.update(schema_result.data)

        return ExecutionResult(data=data, errors=[])
    except GraphQLError as error:
        return ExecutionResult(errors=[error], invalid=True)


class Result(object):
    def __init__(self, value, join_values):
        self.value = value
        self.join_values = join_values


class Value(six.with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def fetch(self, request):
        pass


def field(**kwargs):
    return Field(**kwargs)


class FieldBase(object):
    pass


class Field(FieldBase):
    target = None
    args = {}

    def __init__(self, type, **kwargs):
        self.type = type
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    def to_graphql_field(self):
        return GraphQLField(
            type=self.type,
            resolver=_resolve_fetched_field,
            args=self.args,
        )


def _resolve_fetched_field(source, args, context, info):
    return source[field_key(info.field_asts[0])]


def relationship(join=None, args=None, **kwargs):
    if join is None:
        join = {}
    if args is None:
        args = {}

    return Relationship(join=join, args=args, **kwargs)


class Relationship(FieldBase):
    def __init__(self, target, process_results, wrap_type, build_query, join, args):
        self.target = target
        self.build_query = build_query
        self.join = join
        self.args = args
        self._process_results = process_results
        self._wrap_type = wrap_type

        self._parent_join_keys = tuple("_graphjoiner_joinToChildrenKey_" + parent_key for parent_key in self.join.keys())

    def copy(self, target=None, build_query=None, join=None, args=None):
        if target is None:
            target = self.target
        if build_query is None:
            build_query = self.build_query
        if join is None:
            join = self.join
        if args is None:
            args = self.args

        return Relationship(
            target=target,
            build_query=build_query,
            join=join,
            args=args,
            wrap_type=self._wrap_type,
            process_results=self._process_results,
        )

    def parent_join_selections(self, parent):
        fields = parent.fields()
        return [
            Request(field=fields[field_name], key=key)
            for field_name, key in zip(self.join.keys(), self._parent_join_keys)
        ]

    def fetch(self, request, parent_query):
        query = self.build_query(request.args, parent_query, request.context)
        join_fields = self.target.join_fields()
        join_selections = [
            Request(key="_graphjoiner_joinToParentKey_" + child_key, field=join_fields[child_key])
            for child_key in self.join.values()
        ]

        child_request = assoc(request, join_selections=join_selections)
        results = self.target.fetch(child_request, query)
        return RelationshipResults(
            results=results,
            process_results=self._process_results,
            parent_join_keys=self._parent_join_keys,
        )

    def to_graphql_field(self):
        # TODO: differentiate between root and non-root types properly
        if self.join:
            resolve = _resolve_fetched_field
        else:
            def resolve(source, args, context, info):
                request = request_from_graphql_ast(
                    info.field_asts[0],
                    self.target,
                    context=context,
                    variables=info.variable_values,
                    field=self,
                    fragments=info.fragments,
                )
                return self.fetch(request, None).get(())

        return GraphQLField(
            type=self._wrap_type(self.target.to_graphql_type()),
            resolver=resolve,
            args=self.args,
        )


class RelationshipResults(object):
    def __init__(self, results, process_results, parent_join_keys):
        key_func = lambda result: result.join_values
        self._results = dict(
            (key, [result.value for result in results])
            for key, results in groupby(sorted(results, key=key_func), key=key_func)
        )
        self._process_results = process_results
        self._parent_join_keys = parent_join_keys

    def _parent_join_values(self, parent):
        return tuple(parent[join_field] for join_field in self._parent_join_keys)

    def get(self, key):
        return self._process_results(self._results.get(self._parent_join_values(key), []))


def single(target, build_query, **kwargs):
    return relationship(
        target=target,
        build_query=build_query,
        process_results=_one_or_none,
        wrap_type=lambda graphql_type: graphql_type,
        **kwargs
    )


def _one_or_none(values):
    if len(values) == 0:
        return None
    elif len(values) > 1:
        raise Exception("TODO")
    else:
        return values[0]


def first_or_none(target, build_query, **kwargs):
    return relationship(
        target=target,
        build_query=build_query,
        process_results=_first_or_none,
        wrap_type=lambda graphql_type: graphql_type,
        **kwargs
    )


def _first_or_none(values):
    if len(values) == 0:
        return None
    else:
        return values[0]


def many(target, build_query, **kwargs):
    return relationship(
        target=target,
        build_query=build_query,
        process_results=lambda x: x,
        wrap_type=lambda graphql_type: GraphQLList(graphql_type),
        **kwargs
    )



def extract(relationship, field_name):
    return relationship.copy(
        target=ScalarJoinType(relationship.target, field_name),
    )



class ScalarJoinType(Value):
    def __init__(self, target, field_name):
        self._target = target
        self._field_name = field_name

    @property
    def _field(self):
        return self._target.fields()[self._field_name]

    def fields(self):
        if isinstance(self._field, Relationship):
            return self._field.target.fields()
        else:
            return {}

    def join_fields(self):
        return self._target.join_fields()

    def fetch(self, request, query):
        field_request = Request(key=self._field_name, field=self._field, selections=request.selections, context=request.context)
        results = self._target.fetch(assoc(request, selections=[field_request]), query)
        return [
            Result(value=result.value[self._field_name], join_values=result.join_values)
            for result in results
        ]

    def to_graphql_type(self):
        return self._field.to_graphql_field().type


class JoinType(Value):
    def __init__(self, name, fetch_immediates, fields, interfaces=None):
        if interfaces is None:
            interfaces = ()

        self._name = name
        self._fetch_immediates = fetch_immediates
        self._generate_fields = fields
        self._interfaces = interfaces
        self._fields = None
        self._graphql_type = None

    def fields(self):
        if self._fields is None:
            self._fields = self._generate_fields()
        return self._fields

    def join_fields(self):
        return self.fields()

    def fetch(self, request, query):
        (relationship_selections, requested_immediate_selections) = partition(
            lambda selection: isinstance(selection.field, Relationship),
            request.selections,
        )

        join_to_children_selections = [
            parent_join_selection
            for selection in relationship_selections
            for parent_join_selection in selection.field.parent_join_selections(self)
        ]

        immediate_selections = unique(
            requested_immediate_selections + list(request.join_selections) + join_to_children_selections,
            key=lambda selection: selection.key,
        )


        keys = tuple(selection.key for selection in immediate_selections)

        results = [
            dict(zip(keys, row))
            for row in self._fetch_immediates(immediate_selections, query, request.context)
        ]

        for selection in relationship_selections:
            children = selection.field.fetch(selection, query)
            for result in results:
                result[selection.key] = children.get(result)

        return [
            Result(
                dict((selection.key, result[selection.key]) for selection in request.selections),
                tuple(result[selection.key] for selection in request.join_selections),
            )
            for result in results
        ]

    def to_graphql_type(self):
        if self._graphql_type is None:
            self._graphql_type = GraphQLObjectType(
                name=self._name,
                fields=lambda: dict(
                    (name, field.to_graphql_field())
                    for name, field in six.iteritems(self.fields())
                ),
                interfaces=self._interfaces,
            )

        return self._graphql_type


def RootJoinType(**kwargs):
    return JoinType(fetch_immediates=lambda *_: [()], **kwargs)



import collections
from copy import copy

from attr import attrs, attrib
from graphql.language import ast as ast_types
from graphql.execution.values import get_argument_values
from graphql.type.directives import GraphQLIncludeDirective, GraphQLSkipDirective
from six.moves import filter

from .util import find, single


@attrs
class DocumentRequest(object):
    query = attrib()
    schema_query = attrib()


@attrs
class Request(object):
    key = attrib()
    field = attrib()
    args = attrib(default={})
    selections = attrib(default=[])
    join_selections = attrib(default=[])
    context = attrib(default=None)


def request_from_graphql_document(document, root, context, variables):
    fragments = dict(
        (definition.name.value, definition)
        for definition in document.definitions
        if isinstance(definition, ast_types.FragmentDefinition)
    )
    definition_index, query = single(list(filter(
        lambda pair: isinstance(pair[1], ast_types.OperationDefinition),
        enumerate(document.definitions)
    )))

    schema_selection = find(
        lambda selection: selection.name.value == "__schema",
        query.selection_set.selections,
    )

    if schema_selection is None:
        schema_query = None
    else:
        schema_query_definition = copy(query)
        schema_query_definition.selection_set = copy(schema_query_definition.selection_set)
        schema_query_definition.selection_set.selections = [schema_selection]

        schema_query = copy(document)
        schema_query.definitions = copy(schema_query.definitions)
        schema_query.definitions[definition_index] = schema_query_definition

    return DocumentRequest(
        query=request_from_graphql_ast(query, root, context=context, variables=variables, fragments=fragments, field=None),
        schema_query=schema_query,
    )


def request_from_graphql_ast(ast, root, context, variables, field, fragments):
    if isinstance(ast, ast_types.Field):
        key = field_key(ast)
    else:
        key = None

    if field is None:
        args = {}
    else:
        args = get_argument_values(field.args, getattr(ast, "arguments", []), variables=variables)

    selections = _graphql_selections(ast, root, context=context, variables=variables, fragments=fragments)

    return Request(
        key=key,
        field=field,
        args=args,
        selections=selections,
        context=context,
    )


def _field_name(ast):
    return ast.name.value


def field_key(ast):
    if ast.alias is None:
        return _field_name(ast)
    else:
        return ast.alias.value



def _graphql_selections(ast, root, context, variables, fragments):
    if ast.selection_set:
        fields = root.fields()

        field_selections = _merge_fields(_collect_fields(ast, fragments=fragments, variables=variables))

        return [
            _request_from_selection(
                selection,
                context=context,
                variables=variables,
                fragments=fragments,
                field=fields[_field_name(selection)]
            )
            for selection in field_selections
            if _field_name(selection) != "__schema"
        ]
    else:
        return []


def _collect_fields(ast, fragments, variables):
    field_selections = []

    _add_fields(ast, field_selections, fragments=fragments, variables=variables)

    return field_selections

def _add_fields(ast, field_selections, fragments, variables):
    for selection in ast.selection_set.selections:
        if _should_include_node(selection, variables=variables):
            if isinstance(selection, ast_types.Field):
                field_selections.append(selection)
            elif isinstance(selection, ast_types.FragmentSpread):
                # TODO: handle type conditions
                _add_fields(fragments[selection.name.value], field_selections, fragments=fragments, variables=variables)
            elif isinstance(selection, ast_types.InlineFragment):
                _add_fields(selection, field_selections, fragments=fragments, variables=variables)
            else:
                raise Exception("Unknown selection: {}".format(type(selection)))


def _should_include_node(node, variables):
    for directive in node.directives:
        name = directive.name.value
        if name == "skip":
            args = get_argument_values(GraphQLSkipDirective.args, directive.arguments, variables)
            if args.get("if") is True:
                return False
        elif name == "include":
            args = get_argument_values(GraphQLIncludeDirective.args, directive.arguments, variables)
            if args.get("if") is False:
                return False
        else:
            raise Exception("Unknown directive: {}".format(name))

    return True


def _merge_fields(selections):
    # TODO: validation

    merged = collections.OrderedDict()

    for selection in selections:
        key = field_key(selection)
        if key in merged:
            if selection.selection_set is not None:
                merged[key] = copy(merged[key])
                merged[key].selection_set = copy(merged[key].selection_set)
                merged[key].selection_set.selections += selection.selection_set.selections
        else:
            merged[key] = selection

    return merged.values()


def _request_from_selection(selection, field, context, variables, fragments):
    return request_from_graphql_ast(
        selection,
        context=context,
        variables=variables,
        fragments=fragments,
        field=field,
        root=field.target,
    )




from __future__ import absolute_import

import graphql
import six
import sqlalchemy
from sqlalchemy.orm import Query

from graphjoiner import declarative
from . import field, ObjectType, join_builder


class SqlAlchemyObjectType(ObjectType):
    __abstract__ = True

    @classmethod
    def __primary_key__(cls):
        return cls.__model__.__mapper__.primary_key

    @classmethod
    def __select_all__(cls):
        return Query([]).select_from(cls.__model__)

    @classmethod
    def __fetch_immediates__(cls, selections, query, context):
        query = query.with_entities(*(
            selection.field.column
            for selection in selections
        ))
        for primary_key_column in cls.__primary_key__():
            query = query.add_columns(primary_key_column)

        return query.distinct().with_session(context.session).all()


def column_field(column, type=None):
    if type is None:
        type = _sql_type_to_graphql_type(column.type)
    return field(
        column=column,
        type=type,
    )


@join_builder
def sql_value_join(local, target, join):
    def build_query(parent_query, context):
        return parent_query.with_entities(*(
                local_field.column.label(remote_field.attr_name)
                for local_field, remote_field in six.iteritems(join)
            )) \
            .with_session(context.session) \
            .all()

    join_fields = dict(
        (local_field.field_name, remote_field.field_name)
        for local_field, remote_field in six.iteritems(join)
    )

    return build_query, join_fields


@join_builder
def sql_join(local, target, join=None):
    if join is None:
        local_field_definition, remote_field_definition = _find_foreign_key(local, target)
        local_field = local_field_definition.field()
        remote_field = remote_field_definition.field()
        join = {local_field: remote_field}
    else:
        join = join.copy()

    def build_query(parent_query, context):
        parents = parent_query \
            .with_entities(*(
                local_field.column
                for local_field in join.keys()
            )) \
            .subquery()

        return target.__select_all__() \
            .join(parents, sqlalchemy.and_(
                parent_column == remote_field.column
                for parent_column, remote_field in zip(parents.c.values(), join.values())
            ))

    return build_query, dict(
        (local_field.field_name, remote_field.field_name)
        for local_field, remote_field in join.items()
    )


def _find_foreign_key(local, target):
    foreign_keys = list(_find_join_candidates(local, target))
    if len(foreign_keys) == 1:
        foreign_key, = foreign_keys
        return foreign_key
    else:
        raise Exception("TODO")

def _find_join_candidates(local, target):
    for local_field, target_field in _find_join_candidates_directional(local, target):
        yield local_field, target_field
    for target_field, local_field in _find_join_candidates_directional(target, local):
        yield local_field, target_field

def _find_join_candidates_directional(local, remote):
    for field_definition in _get_simple_field_definitions(local):
        orm_column = field_definition._kwargs["column"]
        if hasattr(orm_column, "property"):
            column, = orm_column.property.columns
            for foreign_key in column.foreign_keys:
                if remote.__model__.__table__ == foreign_key.column.table:
                    remote_primary_key_column, = foreign_key.column.table.primary_key
                    remote_field = _find_field_for_column(remote, remote_primary_key_column)
                    yield field_definition, remote_field


def _find_field_for_column(cls, column):
    for field_definition in _get_simple_field_definitions(cls):
        if field_definition._kwargs["column"] == column:
            return field_definition
    raise Exception("Could not find field in {} for {}".format(cls.__name__, column))

def _get_simple_field_definitions(cls):
    for field_definition in six.itervalues(cls.__dict__):
        if isinstance(field_definition, declarative.SimpleFieldDefinition):
            yield field_definition


_type_mappings = [
    (sqlalchemy.Integer, graphql.GraphQLInt),
    (sqlalchemy.Float, graphql.GraphQLFloat),
    (sqlalchemy.String, graphql.GraphQLString),
    (sqlalchemy.Boolean, graphql.GraphQLBoolean),
]

def _sql_type_to_graphql_type(sql_type):
    for mapped_sql_type, graphql_type in _type_mappings:
        if isinstance(sql_type, mapped_sql_type):
            return graphql_type

    raise Exception("Unknown SQL type: {}".format(sql_type))



#!/usr/bin/env python

import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='graphjoiner',
    version='0.4.0b12',
    description='Implementing GraphQL with joins',
    long_description=read("README.rst"),
    author='Michael Williamson',
    author_email='mike@zwobble.org',
    url='http://github.com/healx/python-graphjoiner',
    packages=['graphjoiner', 'graphjoiner.declarative'],
    keywords="graphql graph join ",
    install_requires=["graphql-core>=1.0.1,<1.1", "attrs>=16.1.0,<17", "six"],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)




from functools import partial
import re

from graphql import GraphQLArgument
import six

import graphjoiner


def executor(root):
    root_type = root.__graphjoiner__
    return graphjoiner.executor(root_type)


class ObjectTypeMeta(type):
    def __new__(meta, name, bases, attrs):
        cls = super(ObjectTypeMeta, meta).__new__(meta, name, bases, attrs)
        if attrs.get("__abstract__"):
            return cls

        def fields():
            return dict(
                (field_definition.field_name, field_definition.__get__(None, cls))
                for key, field_definition in six.iteritems(cls.__dict__)
                if isinstance(field_definition, FieldDefinition)
            )

        cls.__graphjoiner__ = graphjoiner.JoinType(
            name=cls.__name__,
            fields=fields,
            fetch_immediates=cls.__fetch_immediates__,
            interfaces=attrs.get("__interfaces__", []),
        )

        for key, field_definition in six.iteritems(cls.__dict__):
            if isinstance(field_definition, FieldDefinition):
                field_definition.attr_name = key
                field_definition.field_name = _snake_case_to_camel_case(key)
                field_definition._owner = cls

        return cls


class ObjectType(six.with_metaclass(ObjectTypeMeta)):
    __abstract__ = True


class RootType(ObjectType):
    __abstract__ = True

    @staticmethod
    def __fetch_immediates__(*args):
        return [()]


class FieldDefinition(object):
    # TODO: make sure _owner isn't overwritten once set
    _owner = None
    _value = None
    field_name = None
    attr_name = None

    def __get__(self, obj, type=None):
        if self._owner is None:
            self._owner = type

        return self.field()

    def field(self):
        if self._value is None:
            self._value = self.instantiate()
            self._value.field_name = self.field_name
            self._value.attr_name = self.attr_name

        return self._value


def field(**kwargs):
    return SimpleFieldDefinition(**kwargs)


class SimpleFieldDefinition(FieldDefinition):
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def instantiate(self):
        return graphjoiner.field(**self._kwargs)


def join_builder(build_join):
    def wrapped(target, *args, **kwargs):
        filter = kwargs.pop("filter", None)
        return lambda func: RelationshipDefinition(
            func=func,
            target=target,
            filter=filter,
            build_join=lambda local: build_join(local, target, *args, **kwargs),
        )

    return wrapped


def relationship(select_values, relationship_type):
    return LazyFieldDefinition(lambda: select_values()(relationship_type))

first_or_none = partial(relationship, relationship_type=graphjoiner.first_or_none)
single = partial(relationship, relationship_type=graphjoiner.single)
many = partial(relationship, relationship_type=graphjoiner.many)


class RelationshipDefinition(FieldDefinition):
    def __init__(self, func, target, filter, build_join):
        self._func = func
        self._target = target
        self._filter = filter
        self._build_join = build_join
        self._args = []

    def instantiate(self):
        build_query, join = self._build_join(self._owner)

        def build_query_with_args(args, parent_query, context):
            query = build_query(parent_query, context)

            if self._filter is not None:
                query = self._filter(query)

            for arg_name, _, refine_query in self._args:
                if arg_name in args:
                    query = refine_query(query, args[arg_name])

            return query

        args = dict(
            (arg_name, GraphQLArgument(arg_type))
            for arg_name, arg_type, _ in self._args
        )

        return self._func(self._target.__graphjoiner__, build_query_with_args, join=join, args=args)

    def arg(self, arg_name, arg_type):
        def add_arg(refine_query):
            self._args.append((arg_name, arg_type, refine_query))

        return add_arg


def extract(relationship, field_name):
    return ExtractFieldDefinition(relationship, field_name)


class ExtractFieldDefinition(FieldDefinition):
    def __init__(self, relationship, field_name):
        self._relationship = relationship
        self._field_name = field_name

    def instantiate(self):
        if self._relationship._owner is None:
            self._relationship._owner = self._owner

        return graphjoiner.extract(self._relationship.field(), self._field_name)


class LazyFieldDefinition(FieldDefinition):
    def __init__(self, func):
        self._func = func
        self._value = None
        self._setup = []

    def instantiate(self):
        field_definition = self._func()

        for setup in self._setup:
            setup(field_definition)

        field_definition.field_name = self.field_name
        field_definition.attr_name = self.attr_name
        return field_definition.__get__(None, self._owner)

    def arg(self, arg_name, arg_type):
        def add_arg(refine_query):
            self._setup.append(lambda field: field.arg(arg_name, arg_type)(refine_query))

        return add_arg



def _snake_case_to_camel_case(value):
    return value[0].lower() + re.sub(r"_(.)", lambda match: match.group(1).upper(), value[1:])


@join_builder
def select(local, target, join_query=None, join_fields=None):
    if join_fields is None:
        join_fields = {}
    else:
        join_fields = dict(
            (local_field.field_name, remote_field.field_name)
            for local_field, remote_field in six.iteritems(join_fields)
        )

    def build_query(parent_query, context):
        target_query = target.__select_all__()
        if join_query is None:
            return target_query
        else:
            return join_query(parent_query, target_query)

    return build_query, join_fields
